@file:OptIn(ExperimentalContracts::class)

package ru.landrafhomyak.utility.reference_counter

import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlinx.atomicfu.AtomicLong
import kotlinx.atomicfu.atomic
import kotlinx.atomicfu.update
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose1
import ru.landgrafhomyak.utility.highlevel_try_finally.safeAutoClose2

public class CloseableReferenceCounter(private val _errMessage: String) {
	private val _value: AtomicLong = atomic(0L)

	public fun throwClosed() {
		throw IllegalStateException(this._errMessage)
	}

	public fun incref() {
		this._value.update { o ->
			if (o < 0) this.throwClosed()
			return@update o + 1
		}
	}

	public inline fun <R> tryIncref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.incref()
		return safeAutoClose2(onError = this::decref, action = block)
	}

	public fun assertNotClosed() {
		if (this._value.value < 0) this.throwClosed()
	}

	public fun decref() {
		this._value.update(Long::dec)
	}

	public inline fun <R> tryDecref(block: () -> R): R {
		contract {
			callsInPlace(block, InvocationKind.EXACTLY_ONCE)
		}
		this.assertNotClosed()
		return safeAutoClose2(onSuccess = this::decref, action = block)
	}

	public fun close(errExistRefs: String) {
		val state = _CloseableReferenceCounter_LowLevel.compareAndExchange(this._value, 0, _CloseableReferenceCounter_LowLevel.CLOSED_STATE_VALUE)
		when {
			state > 0 -> throw IllegalStateException(errExistRefs)
			state < 0 -> this.throwClosed()
		}
	}

	public inline fun <R> withRef(protected: () -> R): R {
		this.incref()
		return safeAutoClose1(finally = this::decref, action = protected)
	}

	override fun toString(): String {
		val refcntCached = this._value.value
		@Suppress("LiftReturnOrAssignment")
		if (refcntCached < 0)
			return "<ref counter [closed]>"
		else
			return "<ref counter [${refcntCached}]>"
	}
}